/**
 * @file methods/ann/init_rules/init_rules.hpp
 * @author Ryan Curtin
 *
 * Convenience includes for initialization rules for mlpack's neural network
 * toolkit.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_ANN_INIT_RULES_INIT_RULES_HPP
#define MLPACK_METHODS_ANN_INIT_RULES_INIT_RULES_HPP

#include "const_init.hpp"
#include "gaussian_init.hpp"
#include "glorot_init.hpp"
#include "he_init.hpp"
#include "init_rules_traits.hpp"
#include "kathirvalavakumar_subavathi_init.hpp"
#include "lecun_normal_init.hpp"
#include "network_init.hpp"
#include "nguyen_widrow_init.hpp"
#include "oivs_init.hpp"
#include "orthogonal_init.hpp"
#include "random_init.hpp"

#endif
