/**
 * @file methods/reinforcement_learning/environment/environment.hpp
 * @author Ryan Curtin
 *
 * Convenience include for all reinforcement learning environments.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_REINFORCEMENT_LEARNING_ENVIRONMENT_ENVIRONMENT_HPP
#define MLPACK_METHODS_REINFORCEMENT_LEARNING_ENVIRONMENT_ENVIRONMENT_HPP

#include "env_type.hpp"
#include "acrobot.hpp"
#include "cart_pole.hpp"
#include "cont_double_pole_cart.hpp"
#include "continuous_mountain_car.hpp"
#include "double_pole_cart.hpp"
#include "ftn.hpp"
#include "ftn_param.hpp"
#include "mountain_car.hpp"
#include "pendulum.hpp"
#include "reward_clipping.hpp"

#endif
