/**
 * @file methods/reinforcement_learning/worker/worker.hpp
 * @author Ryan Curtin
 *
 * Convenience header to include all workers for reinforcement learning.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_REINFORCEMENT_LEARNING_WORKER_WORKER_HPP
#define MLPACK_METHODS_REINFORCEMENT_LEARNING_WORKER_WORKER_HPP

#include "one_step_q_learning_worker.hpp"
#include "one_step_sarsa_worker.hpp"
#include "n_step_q_learning_worker.hpp"

#endif
