/**
 * @file methods/amf/init_rules/init_rules.hpp
 * @author Ryan Curtin
 *
 * Include all of the initialization rules implemented for AMF.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_AMF_INIT_RULES_INIT_RULES_HPP
#define MLPACK_METHODS_AMF_INIT_RULES_INIT_RULES_HPP

#include "average_init.hpp"
#include "given_init.hpp"
#include "merge_init.hpp"
#include "no_init.hpp"
#include "random_acol_init.hpp"
#include "random_init.hpp"

#endif
