/**
 * @file methods/nmf/nmf.hpp
 * @author Ryan Curtin
 *
 * Convenience include for NMF.  Note that this simply includes the AMF code,
 * where NMF is actually defined.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_NMF_NMF_HPP
#define MLPACK_METHODS_NMF_NMF_HPP

#include <mlpack/methods/amf/amf.hpp>

#endif
