/*

mlpack is a fast, flexible machine learning library, written in C++, that 
aims to provide fast, extensible implementations of cutting-edge machine
learning algorithms. mlpack provides these algorithms as simple command-line
programs, Go bindings, and C++ classes which can then be integrated into 
larger-scale machine learning solutions.

*/
package mlpack // import "mlpack.org/v1/mlpack"
