/**
 * @file approx_kfn.hpp
 * @author Ryan Curtin
 *
 * Include both the QDAFN and DrusillaSelect strategies for approximate furthest
 * neighbor search.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_APPROX_KFN_APPROX_KFN_HPP
#define MLPACK_METHODS_APPROX_KFN_APPROX_KFN_HPP

#include "qdafn.hpp"
#include "drusilla_select.hpp"

#endif
