/**
 * @file core/cv/metrics/metrics.hpp
 * @author Ryan Curtin
 *
 * Convenience include for all metrics for cross-validation.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_CORE_CV_METRICS_METRICS_HPP
#define MLPACK_CORE_CV_METRICS_METRICS_HPP

#include "facilities.hpp"
#include "accuracy.hpp"
#include "average_strategy.hpp"
#include "f1.hpp"
#include "mse.hpp"
#include "precision.hpp"
#include "r2_score.hpp"
#include "recall.hpp"
#include "roc_auc_score.hpp"
#include "silhouette_score.hpp"

#endif
