/**
 * @file methods/decision_tree/fitness_functions/fitness_functions.hpp
 * @author Abhimanyu Dayal
 *
 * Include all FitnessFunction policies for decision trees.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_DECISION_TREE_FITNESS_FUNCTIONS_FITNESS_FUNCTIONS_HPP
#define MLPACK_METHODS_DECISION_TREE_FITNESS_FUNCTIONS_FITNESS_FUNCTIONS_HPP

#include "gini_gain.hpp"
#include "information_gain.hpp"
#include "mad_gain.hpp"
#include "mse_gain.hpp"

#endif
