/**
 * @file core/data/imputation_methods/imputation_methods.hpp
 * @author Ryan Curtin
 *
 * Convenience includes for imputation methods.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_CORE_DATA_IMPUTATION_METHODS_IMPUTATION_METHODS_HPP
#define MLPACK_CORE_DATA_IMPUTATION_METHODS_IMPUTATION_METHODS_HPP

#include "mean_imputation.hpp"
#include "median_imputation.hpp"
#include "listwise_deletion.hpp"
#include "custom_imputation.hpp"

#endif
