/**
 * @file methods/amf/termination_policies/termination_policies.hpp
 * @author Ryan Curtin
 *
 * Include all of the termination policies implemented for AMF.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_AMF_TERMINATION_POLICIES_TERMINATION_POLICIES_HPP
#define MLPACK_METHODS_AMF_TERMINATION_POLICIES_TERMINATION_POLICIES_HPP

#include "complete_incremental_termination.hpp"
#include "incomplete_incremental_termination.hpp"
#include "max_iteration_termination.hpp"
#include "simple_residue_termination.hpp"
#include "simple_tolerance_termination.hpp"
#include "validation_rmse_termination.hpp"

#endif
