/**
 * @file core/metrics/metrics.hpp
 * @author Ryan Curtin
 *
 * Include all performance metrics and scoring functions implemented by mlpack.
 * For distance metrics (e.g. distances in metric spaces that satisfy the
 * triangle inequality and formal definition of distance metric), see
 * core/distances/ instead.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_CORE_METRICS_METRICS_HPP
#define MLPACK_CORE_METRICS_METRICS_HPP

#include "bleu.hpp" // Technically this should go somewhere else...
#include "iou_metric.hpp"
#include "non_maximal_suppression.hpp"

#endif
