/**
 * @file sparse_autoencoder.hpp
 *
 * Convenience include for
 * mlpack/methods/sparse_autoencoder/sparse_autoencoder.hpp.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_SPARSE_AUTOENCODER_HPP
#define MLPACK_SPARSE_AUTOENCODER_HPP

#include "sparse_autoencoder/sparse_autoencoder.hpp"

#endif
