/**
 * @file random_forest.hpp
 *
 * Convenience include for mlpack/methods/random_forest/random_forest.hpp.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_RANDOM_FOREST_HPP
#define MLPACK_RANDOM_FOREST_HPP

#include "random_forest/random_forest.hpp"

#endif
