/**
 * @file core/data/scaler_methods/scaler_methods.hpp
 * @author Ryan Curtin
 *
 * Convenience include for all scaler methods.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_CORE_DATA_SCALER_METHODS_SCALER_METHODS_HPP
#define MLPACK_CORE_DATA_SCALER_METHODS_SCALER_METHODS_HPP

#include "max_abs_scaler.hpp"
#include "mean_normalization.hpp"
#include "min_max_scaler.hpp"
#include "pca_whitening.hpp"
#include "standard_scaler.hpp"
#include "zca_whitening.hpp"

#endif
