/**
 * @file methods/decision_tree/select_functions/select_functions.hpp
 * @author Abhimanyu Dayal
 *
 * Include all selection policies for decision trees.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_DECISION_TREE_SELECT_FUNCTIONS_SELECT_FUNCTIONS_HPP
#define MLPACK_METHODS_DECISION_TREE_SELECT_FUNCTIONS_SELECT_FUNCTIONS_HPP

#include "all_dimension_select.hpp"
#include "mult_random_dimension_select.hpp"
#include "random_dimension_select.hpp"

#endif
