.highlight table td { padding: 5px; }
.highlight table pre { margin: 0; }
.highlight .cm, span.cm { color: #3322ff; font-style: italic; }
.highlight .cp, span.cp { color: #3322ff; font-weight: bold;  }
.highlight .c1, span.c1 { color: #3322ff; font-style: italic; }
.highlight .cs, span.cs { color: #3322ff; font-weight: bold; font-style: italic; }
.highlight .c, .highlight .cd, span.c, span.cd { color: #3322ff; font-style: italic; }
.highlight .err, span.err { color: #a61717; background-color: #e3d2d2; }
.highlight .gd, span.gd { color: #000000; background-color: #ffdddd; }
.highlight .ge, span.ge { color: #000000; font-style: italic; }
.highlight .gr, span.gr { color: #aa0000; }
.highlight .gh, span.gh { color: #999999; }
.highlight .gi, span.gi { color: #000000; background-color: #ddffdd; }
.highlight .go, span.go { color: #888888; }
.highlight .gp, span.gp { color: #555555; }
.highlight .gs, span.gs { font-weight: bold; }
.highlight .gu, span.gu { color: #aaaaaa; }
.highlight .gt, span.gt { color: #aa0000; }
.highlight .kc, span.kc { color: #000000; font-weight: bold; }
.highlight .kd, span.kd { color: #000000; font-weight: bold; }
.highlight .kn, span.kn { color: #000000; font-weight: bold; }
.highlight .kp, span.kp { color: #000000; font-weight: bold; }
.highlight .kr, span.kr { color: #000000; font-weight: bold; }
.highlight .kt, span.kt { color: #449955; font-weight: bold; }
.highlight .k, .highlight .kv, span.k, span.kv { color: #000000; font-weight: bold; }
.highlight .mf, span.mf { color: #009999; }
.highlight .mh, span.mh { color: #009999; }
.highlight .il, span.il { color: #009999; }
.highlight .mi, span.mi { color: #009999; }
.highlight .mo, span.mo { color: #009999; }
.highlight .m, .highlight .mb, .highlight .mx, span.m, span.mb, span.mx { color: #009999; }
.highlight .sb, span.sb { color: #d14; }
.highlight .sc, span.sc { color: #d14; }
.highlight .sd, span.sd { color: #d14; }
.highlight .s2, span.s2 { color: #d14; }
.highlight .se, span.se { color: #d14; }
.highlight .sh, span.sh { color: #d14; }
.highlight .si, span.si { color: #d14; }
.highlight .sx, span.sx { color: #d14; }
.highlight .sr, span.sr { color: #009926; }
.highlight .s1, span.s1 { color: #d14; }
.highlight .ss, span.ss { color: #990073; }
.highlight .s, span.s { color: #d14; }
.highlight .na, span.na { color: #008080; }
.highlight .bp, span.bp { color: #999999; }
.highlight .nb, span.nb { color: #0086B3; }
.highlight .nc, span.nc { color: #449955; font-weight: bold; }
.highlight .no, span.no { color: #008080; }
.highlight .nd, span.nd { color: #3c5d5d; font-weight: bold; }
.highlight .ni, span.ni { color: #800080; }
.highlight .ne, span.ne { color: #990000; font-weight: bold; }
.highlight .nf, span.nf { }
.highlight .nl, span.nl { color: #990000; font-weight: bold; }
.highlight .nn, span.nn { color: #555555; }
.highlight .nt, span.nt { color: #000080; }
.highlight .vc, span.vc { color: #008080; }
.highlight .vg, span.vg { color: #008080; }
.highlight .vi, span.vi { color: #008080; }
.highlight .nv, span.nv { color: #008080; }
.highlight .ow, span.ow { color: #000000; font-weight: bold; }
.highlight .o, span.o { }
.highlight .w, span.w { color: #bbbbbb; }
.highlight {
  background-color: #f8f8f8;
  display: inline-block;
  position: relative;
  left: 50%;
  transform: translateX(-50%);
  max-width: 100%;
  overflow: scroll;
}
