/**
 * @file naive_bayes/naive_bayes.hpp
 * @author Ryan Curtin
 *
 * Convenience include for Naive Bayes classifier code.  This file exists to
 * satisfy the include convention of `module_name/module_name.hpp`.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_NAIVE_BAYES_NAIVE_BAYES_HPP
#define MLPACK_METHODS_NAIVE_BAYES_NAIVE_BAYES_HPP

#include "naive_bayes_classifier.hpp"

#endif
