/**
 * @file methods/pca/decomposition_policies/decomposition_policies.hpp
 * @author Ryan Curtin
 *
 * Convenience include for all PCA decomposition policies.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_PCA_DECOMPOSITION_POLICIES_DECOMPOSITION_POLICIES_HPP
#define MLPACK_METHODS_PCA_DECOMPOSITION_POLICIES_DECOMPOSITION_POLICIES_HPP

#include "exact_svd_method.hpp"
#include "quic_svd_method.hpp"
#include "randomized_block_krylov_method.hpp"
#include "randomized_svd_method.hpp"

#endif
