/**
 * @file methods/decision_tree/split_functions/splits.hpp
 * @author Abhimanyu Dayal
 *
 * Include all split policies for decision trees.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_DECISION_TREE_SPLIT_FUNCTIONS_SPLITS_HPP
#define MLPACK_METHODS_DECISION_TREE_SPLIT_FUNCTIONS_SPLITS_HPP

#include "all_categorical_split.hpp"
#include "best_binary_numeric_split.hpp"
#include "random_binary_numeric_split.hpp"
#include "best_binary_categorical_split.hpp"

#endif
