#' mlpack
#'
#' mlpack is a fast, flexible machine learning library, written in C++, that
#' aims to provide fast, extensible implementations of cutting-edge machine
#' learning algorithms. mlpack provides these algorithms as simple command-line
#' programs, C++ classes and bindings for : Python, Julia, Go and R which can
#' then be integrated into larger-scale machine learning solutions.
#' 
#' @name mlpack
#' @aliases mlpack-package
#' @importFrom Rcpp evalCpp
#' @useDynLib mlpack
"_PACKAGE"
