#!/usr/bin/env python
"""
This is an autogenerated file that allows convenient imports of mlpack
functionality.

mlpack is free software; you may redistribute it and/or modify it under the
terms of the 3-clause BSD license.  You should have received a copy of the
3-clause BSD license along with mlpack.  If not, see
http://www.opensource.org/licenses/BSD-3-Clause for more information.
"""
import warnings
warnings.filterwarnings("default", category=ImportWarning)
