/**
 * @file core/data/string_encoding_policies/string_encoding_policies.hpp
 * @author Ryan Curtin
 *
 * Convenience include for all string encoding policies.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_CORE_DATA_STRING_ENCODING_POLICIES_STRING_ENCODING_POLICIES_HPP
#define MLPACK_CORE_DATA_STRING_ENCODING_POLICIES_STRING_ENCODING_POLICIES_HPP

#include "bag_of_words_encoding_policy.hpp"
#include "dictionary_encoding_policy.hpp"
#include "tf_idf_encoding_policy.hpp"

#include "policy_traits.hpp"

#endif
