/**
 * @file methods/reinforcement_learning/q_networks/q_networks.hpp
 * @author Ryan Curtin
 *
 * Convenience header to include all implemented Q networks.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_REINFORCEMENT_LEARNING_Q_NETWORKS_Q_NETWORKS_HPP
#define MLPACK_METHODS_REINFORCEMENT_LEARNING_Q_NETWORKS_Q_NETWORKS_HPP

#include "simple_dqn.hpp"
#include "categorical_dqn.hpp"
#include "dueling_dqn.hpp"

#endif
