/**
 * @file methods/cf/normalization/normalization.hpp
 * @author Ryan Curtin
 *
 * Convenience include for all normalization policies implemented for CF.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_CF_NORMALIZATION_NORMALIZATION_HPP
#define MLPACK_METHODS_CF_NORMALIZATION_NORMALIZATION_HPP

#include "combined_normalization.hpp"
#include "item_mean_normalization.hpp"
#include "no_normalization.hpp"
#include "overall_mean_normalization.hpp"
#include "user_mean_normalization.hpp"
#include "z_score_normalization.hpp"

#endif
