/**
 * @file methods/amf/update_rules/update_rules.hpp
 * @author Ryan Curtin
 *
 * Include all of the update rules implemented for AMF.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_AMF_UPDATE_RULES_UPDATE_RULES_HPP
#define MLPACK_METHODS_AMF_UPDATE_RULES_UPDATE_RULES_HPP

#include "nmf_mult_dist.hpp"
#include "nmf_mult_div.hpp"
#include "nmf_als.hpp"
#include "svd_batch_learning.hpp"
#include "svd_incomplete_incremental_learning.hpp"
#include "svd_complete_incremental_learning.hpp"

#endif
