/**
 * @file tests/kde_model_test.cpp
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */

// The test as of now only ensures that the header is self-contained.
// As soon as more useful tests are added, it should still be ensured
// that the main header is included first to preserve the self-containedness.
#include <mlpack/methods/kde/kde_model.hpp>
